# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import multiprocessing
import unittest
import unittest.mock as mock
from multiprocessing import Queue, Value
from multiprocessing.connection import Pipe

import cloudpickle
from ai_flow.meta.project_meta import ProjectMeta

from ai_flow.workflow.status import Status

from ai_flow.api.context_extractor import ContextExtractor, EventContext, Broadcast, ContextList

from ai_flow.meta.workflow_meta import WorkflowMeta
from ai_flow.scheduler_service.service.workflow_event_handler import WorkflowEventHandler
from ai_flow.scheduler_service.service.workflow_execution_event_handler_state import WorkflowContextEventHandlerState
from ai_flow.workflow.control_edge import WorkflowSchedulingRule, MeetAllEventCondition, WorkflowAction

from notification_service.base_notification import BaseEvent

from ai_flow.plugin_interface.scheduler_interface import Scheduler, WorkflowExecutionInfo

from ai_flow.store.abstract_store import AbstractStore

from ai_flow.scheduler_service.service.workflow_event_processor import WorkflowEventProcessor, Poison


class MyContextExtractor(ContextExtractor):

    def extract_context(self, event: BaseEvent) -> EventContext:
        if event.event_type == 'exception':
            raise Exception()
        if event.event_type == 'broadcast':
            return Broadcast()

        context_list = ContextList()
        context_list.add_context(event.context)
        return context_list


class TestWorkflowEventProcessor(unittest.TestCase):

    def setUp(self) -> None:
        self.mock_store: AbstractStore = mock.Mock()
        self.mock_scheduler: Scheduler = mock.Mock()
        self.mock_event_handler: WorkflowEventHandler = mock.Mock()

        def mock_event_handler_factory(scheduler_rule):
            return self.mock_event_handler

        self.c1, self.c2 = Pipe()
        self.processor = WorkflowEventProcessor(self.c1, self.mock_store, self.mock_scheduler,
                                                workflow_event_handler_factory=mock_event_handler_factory)

        self._prepare_workflows()

    def test_run_and_stop(self):
        import time
        self.call_cnt = Value('i', 0)

        def mock__process_event(*args, **kwargs):
            self.call_cnt.value += 1

        self.processor._process_event = mock__process_event
        process = multiprocessing.Process(target=self.processor.run)
        process.start()

        event = BaseEvent('k', 'v', namespace='test_namespace')
        self.c2.send(event)

        time.sleep(1)
        self.assertEqual(1, self.call_cnt.value)

        self.c2.send(Poison())

        process.join()

    def test__process_event(self):
        class WorkflowMetaMatcher:
            def __init__(self, project_id, name):
                self.project_id = project_id
                self.name = name

            def __eq__(self, other):
                if not isinstance(other, WorkflowMeta):
                    return False
                return self.project_id == other.project_id and self.name == other.name

        with mock.patch.object(self.processor, '_handle_event_for_workflow') as handle_method, \
                mock.patch.object(self.processor, '_update_workflow_last_event_version') as update_last_event_version:
            e = BaseEvent('k', 'v', namespace='test_namespace')
            self.processor._process_event(e)
            handle_method.assert_has_calls([mock.call('test_project1', WorkflowMetaMatcher(0, 'workflow1'), e),
                                            mock.call('test_project2', WorkflowMetaMatcher(1, 'workflow2'), e),
                                            mock.call('test_project2', WorkflowMetaMatcher(1, 'workflow3'), e)])
            update_last_event_version.assert_has_calls(
                [mock.call('test_project1', WorkflowMetaMatcher(0, 'workflow1'), e),
                 mock.call('test_project2', WorkflowMetaMatcher(1, 'workflow2'), e),
                 mock.call('test_project2', WorkflowMetaMatcher(1, 'workflow3'), e)])

    def _prepare_workflows(self):
        context_extractor = MyContextExtractor()

        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.STOP)
        rule1 = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k1', 'v1', namespace='test_namespace'),
                                       WorkflowAction.START)
        rule2 = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k2', 'v2', namespace='test_namespace'),
                                       WorkflowAction.START)
        w1 = WorkflowMeta('workflow1', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule, rule1])
        w2 = WorkflowMeta('workflow2', 1, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule, rule2])
        w3 = WorkflowMeta('workflow3', 1, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule])

        def mock_list_workflows(project_name):
            if project_name == 'test_project1':
                return [w1]
            elif project_name == 'test_project2':
                return [w2, w3]
            else:
                return None

        self.mock_store.list_workflows = mock_list_workflows
        p1 = ProjectMeta(name='test_project1', uri='dummy')
        p2 = ProjectMeta(name='test_project2', uri='dummy')
        self.mock_store.list_projects.return_value = [p1, p2]

    def test__get_workflow_execution_state_register_state_if_not_exist(self):
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1')
        self.mock_store.get_workflow_context_event_handler_state.return_value = None
        self.mock_store.register_workflow_context_event_handler_state.return_value = state

        context_list = ContextList()
        context_list.add_context('context_1')
        states = self.processor._get_workflow_execution_state(context_list, 'project', 'workflow1')

        self.assertEqual(1, len(states))
        self.assertEqual(state, states[0])
        self.mock_store.register_workflow_context_event_handler_state.assert_called_with('project', 'workflow1',
                                                                                         'context_1')

    def test__get_workflow_execution_state_with_context(self):
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1')
        self.mock_store.get_workflow_context_event_handler_state.return_value = state

        context_list = ContextList()
        context_list.add_context('context_1')
        context_list.add_context('context_2')
        states = self.processor._get_workflow_execution_state(context_list, 'project', 'workflow1')

        calls = [mock.call('project', 'workflow1', 'context_1'), mock.call('project', 'workflow1', 'context_2')]
        self.mock_store.get_workflow_context_event_handler_state.assert_has_calls(calls, any_order=True)
        self.assertEqual(2, len(states))

    def test__get_workflow_execution_state_with_broadcast(self):
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1')
        self.mock_store.list_workflow_context_event_handler_states.return_value = [state]

        states = self.processor._get_workflow_execution_state(Broadcast(), 'project', 'workflow1')

        self.mock_store.list_workflow_context_event_handler_states.assert_called_with('project', 'workflow1')
        self.assertEqual(1, len(states))
        self.assertEqual(state, states[0])

    def test__handler_event_for_workflow_none_action(self):
        context_extractor = MyContextExtractor()
        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.START)
        rule1 = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k1', 'v1', namespace='test_namespace'),
                                       WorkflowAction.START)
        w1 = WorkflowMeta('workflow1', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule, rule1])
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1')
        self.mock_store.get_workflow_context_event_handler_state.return_value = state
        self.mock_event_handler.handle_event.return_value = (WorkflowAction.NONE, 1)

        e = BaseEvent('k1', 'v1', namespace='test_namespace')
        self.processor._handle_event_for_workflow('project', w1, e)

        self.mock_store.update_workflow_context_event_handler_state \
            .assert_called_with('project', 'workflow1', 'context_1', None, 1)

    def test__handler_event_for_workflow_start_action(self):
        context_extractor = MyContextExtractor()
        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.START)
        rule1 = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k1', 'v1', namespace='test_namespace'),
                                       WorkflowAction.START)
        w1 = WorkflowMeta('workflow1', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule, rule1])
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1')
        self.mock_store.get_workflow_context_event_handler_state.return_value = state

        # Start Action
        self.mock_scheduler.start_new_workflow_execution.return_value = WorkflowExecutionInfo('execution_id')
        self.mock_event_handler.handle_event.return_value = (WorkflowAction.START, 1)
        e = BaseEvent('k1', 'v1', namespace='test_namespace')
        self.processor._handle_event_for_workflow('project', w1, e)

        self.mock_scheduler.start_new_workflow_execution.assert_called_with('project', 'workflow1', 'context_1')
        self.mock_store.update_workflow_context_event_handler_state \
            .assert_called_with('project', 'workflow1', 'context_1', 'execution_id', 1)

    def test__handler_event_for_workflow_start_with_running_workflow_execution(self):
        context_extractor = MyContextExtractor()
        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.START)
        rule1 = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k1', 'v1', namespace='test_namespace'),
                                       WorkflowAction.START)
        w1 = WorkflowMeta('workflow1', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule, rule1])
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1', '1')
        self.mock_store.get_workflow_context_event_handler_state.return_value = state

        # Start Action
        self.mock_scheduler.get_workflow_execution.return_value = WorkflowExecutionInfo('1', status=Status.RUNNING)
        self.mock_event_handler.handle_event.return_value = (WorkflowAction.START, 1)
        e = BaseEvent('k1', 'v1', namespace='test_namespace')
        self.processor._handle_event_for_workflow('project', w1, e)

        self.mock_scheduler.start_new_workflow_execution.assert_not_called()
        self.mock_store.update_workflow_context_event_handler_state \
            .assert_called_with('project', 'workflow1', 'context_1', '1', 1)

    def test__handler_event_for_workflow_start_with_non_running_workflow_execution(self):
        context_extractor = MyContextExtractor()
        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.START)
        rule1 = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k1', 'v1', namespace='test_namespace'),
                                       WorkflowAction.START)
        w1 = WorkflowMeta('workflow1', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule, rule1])
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1', '1')
        self.mock_store.get_workflow_context_event_handler_state.return_value = state

        # Start Action
        self.mock_scheduler.get_workflow_execution.return_value = WorkflowExecutionInfo('1', status=Status.FINISHED)
        self.mock_scheduler.start_new_workflow_execution.return_value = WorkflowExecutionInfo('execution_id')
        self.mock_event_handler.handle_event.return_value = (WorkflowAction.START, 1)
        e = BaseEvent('k1', 'v1', namespace='test_namespace')
        self.processor._handle_event_for_workflow('project', w1, e)

        self.mock_scheduler.start_new_workflow_execution.assert_called_with('project', 'workflow1', 'context_1')
        self.mock_store.update_workflow_context_event_handler_state \
            .assert_called_with('project', 'workflow1', 'context_1', 'execution_id', 1)

    def test__handler_event_for_workflow_stop_action(self):
        context_extractor = MyContextExtractor()
        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.START)
        rule1 = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k1', 'v1', namespace='test_namespace'),
                                       WorkflowAction.START)
        w1 = WorkflowMeta('workflow1', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule, rule1])
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1')
        self.mock_store.get_workflow_context_event_handler_state.return_value = state
        self.mock_event_handler.handle_event.return_value = (WorkflowAction.STOP, 1)

        e = BaseEvent('k1', 'v1', namespace='test_namespace')
        self.processor._handle_event_for_workflow('project', w1, e)

        self.mock_store.update_workflow_context_event_handler_state \
            .assert_called_with('project', 'workflow1', 'context_1', None, 1)

        state.workflow_execution_id = 'execution_id'
        self.processor._handle_event_for_workflow('project', w1, e)
        self.mock_scheduler.stop_workflow_execution.assert_called_with('execution_id')
        self.mock_store.update_workflow_context_event_handler_state \
            .assert_called_with('project', 'workflow1', 'context_1', 'execution_id', 1)

    def test__handler_event_for_workflow_exception(self):
        context_extractor = MyContextExtractor()
        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.START)
        rule1 = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k1', 'v1', namespace='test_namespace'),
                                       WorkflowAction.START)
        w1 = WorkflowMeta('workflow1', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                          scheduling_rules=[rule, rule1])
        state = WorkflowContextEventHandlerState('project', 'workflow1', 'context_1')
        self.mock_store.get_workflow_context_event_handler_state.return_value = state

        # Start Action
        e = BaseEvent('k1', 'v1', namespace='test_namespace', event_type='exception')
        self.processor._handle_event_for_workflow('project', w1, e)

    def test__update_workflow_last_event_version(self):
        e = BaseEvent('k1', 'v1', namespace='test_project', version=15213)
        workflow = WorkflowMeta('w', 0)
        self.processor._update_workflow_last_event_version('test_project', workflow, e)
        self.mock_store.update_workflow.assert_called_once_with('w', 'test_project', context_extractor_in_bytes=None,
                                                                scheduling_rules=[], last_event_version=15213)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# 'License'); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Workflow command"""
import os
import sys

from ai_flow.api.workflow_operation import delete_workflow, get_workflow, get_workflow_execution, list_workflows, \
    list_workflow_executions, pause_workflow_scheduling, resume_workflow_scheduling, start_new_workflow_execution, \
    stop_all_workflow_executions, stop_workflow_execution, submit_workflow
from ai_flow.cli.simple_table import AIFlowConsole
from ai_flow.common.module_load import load_module
from ai_flow.util.cli_utils import init_config
from ai_flow.util.time_utils import parse_date


@init_config
def workflow_delete(args):
    """Deletes all DB records related to the specified workflow."""
    if (
            args.yes
            or input('This will drop all existing records related to the specified workflow. Proceed? (y/n)').upper()
            == 'Y'
    ):
        workflow = delete_workflow(workflow_name=args.workflow_name)
        print("Workflow: {}, deleted: {}.".format(args.workflow_name, workflow is not None))
    else:
        print('Cancelled')


@init_config
def workflow_list(args):
    """Lists all the workflows."""
    workflows = list_workflows(sys.maxsize, 0)
    AIFlowConsole().print_as(
        data=sorted(workflows, key=lambda w: w.workflow_name),
        output=args.output,
        mapper=lambda x: {
            'namespace': x.namespace,
            'workflow_name': x.workflow_name,
            'properties': x.properties,
            'scheduling_rules': x.scheduling_rules
        },
    )


@init_config
def workflow_list_executions(args):
    """Lists all workflow executions of the workflow by workflow name."""
    workflow_executions = list_workflow_executions(args.workflow_name)
    AIFlowConsole().print_as(
        data=sorted(workflow_executions, key=lambda w: w.workflow_execution_id),
        output=args.output,
        mapper=lambda x: {
            'workflow_execution_id': x.workflow_execution_id,
            'workflow_name': x.workflow_info.workflow_name,
            'status': x.status,
            'properties': x.properties,
            'start_date': parse_date(x.start_date),
            'end_date': parse_date(x.end_date),
            'context': x.context
        },
    )


@init_config
def workflow_pause_scheduling(args):
    """Pauses a workflow scheduling."""
    workflow = pause_workflow_scheduling(args.workflow_name)
    print("Workflow: {}, paused: {}.".format(args.workflow_name, workflow is not None))


@init_config
def workflow_resume_scheduling(args):
    """Resumes a paused workflow scheduling."""
    workflow = resume_workflow_scheduling(args.workflow_name)
    print("Workflow: {}, resumed: {}.".format(args.workflow_name, workflow is not None))


@init_config
def workflow_show(args):
    """Shows the workflow by workflow name."""
    workflow = get_workflow(args.workflow_name)
    AIFlowConsole().print_as(
        data=[workflow],
        output=args.output,
        mapper=lambda x: {
            'namespace': x.namespace,
            'workflow_name': x.workflow_name,
            'properties': x.properties,
            'scheduling_rules': x.scheduling_rules
        },
    )


@init_config
def workflow_show_execution(args):
    """Shows the workflow execution by workflow execution id."""
    workflow_execution = get_workflow_execution(args.workflow_execution_id)
    AIFlowConsole().print_as(
        data=[workflow_execution],
        output=args.output,
        mapper=lambda x: {
            'workflow_execution_id': x.workflow_execution_id,
            'workflow_name': x.workflow_info.workflow_name,
            'status': x.status,
            'properties': x.properties,
            'start_date': parse_date(x.start_date),
            'end_date': parse_date(x.end_date),
            'context': x.context
        },
    )


@init_config
def workflow_start_execution(args):
    """Starts a new workflow execution by workflow name."""
    workflow_execution = start_new_workflow_execution(args.workflow_name, args.context)
    print("Workflow: {}, started: {}.".format(args.workflow_name, workflow_execution is not None))


@init_config
def workflow_stop_execution(args):
    """Stops the workflow execution by workflow execution id."""
    workflow_execution = stop_workflow_execution(args.workflow_execution_id)
    print("Workflow Execution: {}, stopped: {}.".format(args.workflow_execution_id, workflow_execution is not None))


@init_config
def workflow_stop_executions(args):
    """Stops all workflow executions by workflow name."""
    workflow_executions = stop_all_workflow_executions(args.workflow_name)
    print("Workflow: {}, stopped: {}.".format(args.workflow_name, str(len(workflow_executions) > 0)))


@init_config
def workflow_submit(args):
    """Submits the workflow by workflow name."""
    dependencies_path = os.path.join(args.project_path, 'dependencies', 'python')
    if os.path.exists(dependencies_path):
        sys.path.insert(0, dependencies_path)
    workflow_path = os.path.join(args.project_path, 'workflows', args.workflow_name)
    sys.path.insert(0, workflow_path)
    workflow_file_path = os.path.join(args.project_path,
                                      'workflows',
                                      args.workflow_name,
                                      '{}.py'.format(args.workflow_name))
    load_module(workflow_file_path)
    workflow = submit_workflow(args.workflow_name)
    print("Workflow: {}, submitted: {}.".format(args.workflow_name, workflow is not None))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
from abc import ABC, abstractmethod
from typing import Text, Dict

from ai_flow.common.configuration import AIFlowConfiguration
from ai_flow.common.module_load import import_string


class BlobConfig(AIFlowConfiguration):

    def __init__(self, config: Dict):
        super().__init__()
        if config is None:
            raise Exception(
                'The `{}` option is not configured in the {} option. Please add it!'.format('blob', 'project.yaml'))
        self['blob_manager_class'] = None
        if config.get('blob_manager_class') is None:
            raise Exception(
                'The `blob_manager_class` option of blob config is not configured. '
                'Please add the `blob_manager_class` option under the `blob` option!')
        self['blob_manager_class'] = config.get('blob_manager_class')
        self['blob_manager_config'] = {}
        if config.get('blob_manager_config') is not None:
            self['blob_manager_config'] = config.get('blob_manager_config')


    def blob_manager_class(self):
        return self.get('blob_manager_class')

    def set_blob_manager_class(self, value):
        self['blob_manager_class'] = value

    def blob_manager_config(self):
        return self['blob_manager_config']

    def set_blob_manager_config(self, value):
        self['blob_manager_config'] = value


class BlobManager(ABC):
    """
    BlobManager is responsible for uploading and downloading files and resources for ai-flow projects.
    Every project should have its own BlobManager to store resources, all resources are uploaded to
    root directory of BlobManager.

    """
    def __init__(self, config: Dict):
        self._log = logging.getLogger(self.__class__.__module__ + '.' + self.__class__.__name__)
        self.config = config
        self.root_dir = config.get('root_directory')

    @property
    def log(self) -> logging.Logger:
        return self._log

    @abstractmethod
    def upload(self, local_file_path: Text) -> Text:
        """
        Upload a given file to blob server. Uploaded file will be placed under self.root_dir.

        :param local_file_path: the path of file to be uploaded.
        :return the uri of the uploaded file in blob server.
        """
        pass

    @abstractmethod
    def download(self, remote_file_path: Text, local_dir: Text) -> Text:
        """
        Download file from remote blob server to local directory.
        Only files located in self.root_dir can be downloaded by BlobManager.

        :param remote_file_path: The path of file to be downloaded.
        :param local_dir: the local directory.
        :return the local uri of the downloaded file.
        """
        pass


class BlobManagerFactory:
    @classmethod
    def create_blob_manager(cls, class_name, config: Dict) -> BlobManager:
        """
        :param class_name: The class name of a (~class:`ai_flow.plugin_interface.blob_manager_interface.BlobManager`)
        :param config: The configuration of the BlobManager.
        """
        if class_name is None:
            return None
        class_object = import_string(class_name)
        return class_object(config)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import abstractmethod
from typing import Dict, Text, Tuple

from pyflink.datastream import StreamExecutionEnvironment
from pyflink.version import __version__

from ai_flow.context.job_context import current_job_name
from ai_flow_plugins.job_plugins.flink.flink_version import INCLUDE_DATASET_VERSIONS
from ai_flow_plugins.job_plugins.flink.flink_wrapped_env import WrappedStatementSet, WrappedTableEnvironment

if __version__ in INCLUDE_DATASET_VERSIONS:
    from pyflink.dataset import ExecutionEnvironment
    from pyflink.table import (BatchTableEnvironment, StreamTableEnvironment, TableConfig)

    from ai_flow_plugins.job_plugins.flink.flink_wrapped_env import (
        WrappedBatchTableEnvironment,
        WrappedStreamTableEnvironment)
else:
    from pyflink.table import (TableEnvironment, EnvironmentSettings)


class FlinkEnv(object):
    """
    FlinkEnv is responsible for creating the objects(ExecutionEnvironment, TableEnvironment, StatementSet)
    needed to build a flink job.
    """
    if __version__ in INCLUDE_DATASET_VERSIONS:
        @abstractmethod
        def create_env(self) -> Tuple[StreamExecutionEnvironment, ExecutionEnvironment, WrappedTableEnvironment,
                                      WrappedStatementSet]:
            pass
    else:
        @abstractmethod
        def create_env(self) -> Tuple[StreamExecutionEnvironment, WrappedTableEnvironment, WrappedStatementSet]:
            pass


class FlinkBatchEnv(FlinkEnv):
    """
    FlinkBatchEnv is the default implementation of FlinkEnv, used in flink batch jobs.
    """
    if __version__ in INCLUDE_DATASET_VERSIONS:
        def create_env(self) -> Tuple[StreamExecutionEnvironment, ExecutionEnvironment, WrappedTableEnvironment,
                                      WrappedStatementSet]:
            exec_env = ExecutionEnvironment.get_execution_environment()
            exec_env.set_parallelism(1)
            _t_env = BatchTableEnvironment.create(exec_env, TableConfig())
            t_env = WrappedBatchTableEnvironment.create_from(_t_env)
            t_env.get_config().get_configuration().set_string("taskmanager.memory.task.off-heap.size", '80m')
            statement_set = t_env.create_statement_set()
            return None, exec_env, t_env, statement_set
    else:
        def create_env(self) -> Tuple[StreamExecutionEnvironment, WrappedTableEnvironment, WrappedStatementSet]:
            exec_env = StreamExecutionEnvironment.get_execution_environment()
            exec_env.set_parallelism(1)
            _t_env = TableEnvironment.create(EnvironmentSettings.in_batch_mode())
            t_env = WrappedTableEnvironment.create_from(_t_env)
            t_env.get_config().get_configuration().set_string("taskmanager.memory.task.off-heap.size", '80m')
            statement_set = t_env.create_statement_set()
            return exec_env, t_env, statement_set


class FlinkStreamEnv(FlinkEnv):
    """
    FlinkStreamEnv is the default implementation of FlinkEnv, used in flink streaming jobs.
    """
    if __version__ in INCLUDE_DATASET_VERSIONS:
        def create_env(self) -> Tuple[StreamExecutionEnvironment, ExecutionEnvironment, WrappedTableEnvironment,
                                      WrappedStatementSet]:
            stream_exec_env: StreamExecutionEnvironment = StreamExecutionEnvironment.get_execution_environment()
            stream_exec_env.set_parallelism(1)
            _t_env = StreamTableEnvironment.create(stream_exec_env, TableConfig())
            t_env = WrappedStreamTableEnvironment.create_from(_t_env)
            t_env.get_config().get_configuration().set_string("taskmanager.memory.task.off-heap.size", '80m')
            statement_set = t_env.create_statement_set()
            return stream_exec_env, None, t_env, statement_set
    else:
        def create_env(self) -> Tuple[StreamExecutionEnvironment, WrappedTableEnvironment, WrappedStatementSet]:
            stream_exec_env: StreamExecutionEnvironment = StreamExecutionEnvironment.get_execution_environment()
            stream_exec_env.set_parallelism(1)
            _t_env = TableEnvironment.create(EnvironmentSettings.in_streaming_mode())
            t_env = WrappedTableEnvironment.create_from(_t_env)
            t_env.get_config().get_configuration().set_string("taskmanager.memory.task.off-heap.size", '80m')
            statement_set = t_env.create_statement_set()
            return stream_exec_env, t_env, statement_set


__flink_env__: FlinkEnv = FlinkBatchEnv()

__job_flink_env_dict__: Dict[Text, FlinkEnv] = {}


def set_flink_env(env: FlinkEnv):
    global __flink_env__, __job_flink_env_dict__
    if current_job_name() is None:
        __flink_env__ = env
    else:
        __job_flink_env_dict__[current_job_name()] = env


def get_global_flink_env() -> FlinkEnv:
    global __flink_env__
    return __flink_env__


def get_flink_env_by_job_name(job_name: Text) -> FlinkEnv:
    global __job_flink_env_dict__
    return __job_flink_env_dict__.get(job_name)



# coding:utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import subprocess
import sys
from shutil import copytree, rmtree
from setuptools import setup, find_packages
from typing import Dict, List

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
in_source = os.path.isfile(CURRENT_DIR + "/run_tests.sh")

devel = [
    'coverage>=6.1.1',
    'flake8',
    'pytest',
    'mock',
]
mysql = [
    'pymysql==0.9.3',
    'mysqlclient>=1.3.6,<1.4',
]
celery = [
    'apache-airflow-providers-celery==1.0.1',
    'redis~=3.2',
]
hdfs = [
    'hdfs~=2.6.0',
]
oss = [
    'oss2==2.9.1',
]
s3 = [
    'boto3~=1.19.7',
    'botocore~=1.22.7',
]
mongo = [
    'mongoengine~=0.22.1',
]
example_requires = [
    'joblib==1.0.1',
    'numpy==1.18.1',
    'pandas==0.24.2',
    'scikit-learn==0.21.2',
]
flink = [
    'apache-flink==1.12.5',
]

devel = devel + mongo + mysql + oss + s3 + hdfs + flink

EXTRAS_REQUIREMENTS: Dict[str, List[str]] = {
    'devel': devel,
    'mysql': mysql,
    'celery': celery,
    'hdfs': hdfs,
    'oss': oss,
    'mongo': mongo,
    'example_requires': example_requires,
}


def remove_if_exists(file_path):
    if os.path.exists(file_path):
        if os.path.islink(file_path) or os.path.isfile(file_path):
            os.remove(file_path)
        else:
            assert os.path.isdir(file_path)
            rmtree(file_path)


def remove_installed_airflow():
    from distutils.sysconfig import get_python_lib

    local_site_package = get_python_lib()
    installed_airflow_path = os.path.join(local_site_package, "airflow")
    for file in os.listdir(installed_airflow_path):
        abs_path = os.path.join(installed_airflow_path, file)
        if os.path.isdir(abs_path) and file == 'providers':
            print("Airflow providers are not being removed.")
        else:
            remove_if_exists(abs_path)

def compile_frontend():  # noqa
    # """Run a command to compile and build aiflow frontend."""
    subprocess.check_call('./ai_flow/frontend/compile_frontend.sh')


def compile_assets():  # noqa
    # """Run a command to compile and build airflow assets."""
    subprocess.check_call('./lib/airflow/airflow/www/compile_assets.sh')


def get_script():
    bin_dir = os.path.join(CURRENT_DIR, "bin")
    return [os.path.join("bin", filename) for filename in os.listdir(bin_dir)]


version_file = os.path.join(CURRENT_DIR, 'ai_flow/version.py')
try:
    exec(open(version_file).read())
except IOError:
    print("Failed to load ai_flow version file for packaging. " +
          "'%s' not found!" % version_file,
          file=sys.stderr)
    sys.exit(-1)
VERSION = __version__ # noqa


try:
    if in_source:
        if os.getenv('INSTALL_AIRFLOW_WITHOUT_FRONTEND') != 'true':
            compile_assets()
        AIRFLOW_DIR = CURRENT_DIR + "/lib/airflow"
        try:
            os.symlink(AIRFLOW_DIR + "/airflow", CURRENT_DIR + "/airflow")
        except BaseException:  # pylint: disable=broad-except
            copytree(AIRFLOW_DIR + "/airflow", CURRENT_DIR + "/airflow")
    else:
        remove_installed_airflow()

    if os.getenv('INSTALL_AIFLOW_WITHOUT_FRONTEND') != 'true':
        compile_frontend()

    require_file = '{}/{}'.format(os.path.dirname(os.path.abspath(__file__)), "requirements.txt")
    with open(require_file) as f:
        context = f.read()
        require_file_lines = context.strip().split('\n')
    require_packages = ['notification-service=={}'.format(VERSION)]

    for line in require_file_lines:
        if os.getenv('BUILD_MINI_AI_FLOW_PACKAGE') == 'true' and line.startswith("# Optional"):
            break
        if not len(line.strip()) == 0 and not line.startswith("#"):
            require_packages.append(line)

    packages = find_packages()
    setup(
        name='ai_flow',
        version=VERSION,
        description='An open source framework that bridges big data and AI.',
        author='',
        author_email='flink.aiflow@gmail.com',
        url='https://github.com/flink-extended/ai-flow',
        packages=find_packages(),
        install_requires=require_packages,
        extras_require=EXTRAS_REQUIREMENTS,
        python_requires='>=3.6, <3.8' if os.getenv('BUILD_MINI_AI_FLOW_PACKAGE') == 'true' else '>=3.7, <3.8',
        include_package_data=True,
        scripts=get_script(),
        package_data={'ai_flow': ['alembic.ini']},
        entry_points={
            'console_scripts': [
                'aiflow = ai_flow.__main__:main'
            ]
        }
    )
finally:
    if in_source:
        remove_if_exists(CURRENT_DIR + "/airflow")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Standard task runner"""
import logging
import os

import psutil
from setproctitle import setproctitle  # pylint: disable=no-name-in-module

from airflow.settings import CAN_FORK
from airflow.task.task_runner.base_task_runner import BaseTaskRunner
from airflow.utils.process_utils import reap_process_group


class StandardTaskRunner(BaseTaskRunner):
    """Standard runner for all tasks."""

    def __init__(self, local_task_job):
        super().__init__(local_task_job)
        self._rc = None
        self.dag = local_task_job.task_instance.task.dag

    def start(self):
        if CAN_FORK and not self.run_as_user:
            self.process = self._start_by_fork()
        else:
            self.process = self._start_by_exec()

    def _start_by_exec(self):
        subprocess = self.run_command()
        return psutil.Process(subprocess.pid)

    def _start_by_fork(self):  # pylint: disable=inconsistent-return-statements
        pid = os.fork()
        if pid:
            self.log.info("Started process %d to run task", pid)
            return psutil.Process(pid)
        else:
            import signal

            from airflow import settings
            from airflow.cli.cli_parser import get_parser
            from airflow.sentry import Sentry

            signal.signal(signal.SIGINT, signal.SIG_DFL)
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
            # Start a new process group
            os.setpgid(0, 0)

            # Force a new SQLAlchemy session. We can't share open DB handles
            # between process. The cli code will re-create this as part of its
            # normal startup
            settings.engine.pool.dispose()
            settings.engine.dispose()

            parser = get_parser()
            # [1:] - remove "airflow" from the start of the command
            args = parser.parse_args(self._command[1:])

            self.log.info('Running: %s', self._command)
            self.log.info('Job %s: Subtask %s', self._task_instance.job_id, self._task_instance.task_id)

            proc_title = "airflow task runner: {0.dag_id} {0.task_id} {0.execution_date}"
            if hasattr(args, "job_id"):
                proc_title += " {0.job_id}"
            setproctitle(proc_title.format(args))

            try:
                args.func(args, dag=self.dag)
                return_code = 0
            except Exception as e:  # pylint: disable=broad-except
                self.log.error("Task runner failed", exc_info=e)
                return_code = 1
            finally:
                # Explicitly flush any pending exception to Sentry if enabled
                Sentry.flush()
                logging.shutdown()
                os._exit(return_code)  # pylint: disable=protected-access

    def return_code(self, timeout=0):
        # We call this multiple times, but we can only wait on the process once
        if self._rc is not None or not self.process:
            return self._rc

        try:
            self._rc = self.process.wait(timeout=timeout)
            self.process = None
        except psutil.TimeoutExpired:
            pass

        return self._rc

    def terminate(self):
        if self.process is None:
            return

        # Reap the child process - it may already be finished
        _ = self.return_code(timeout=0)

        if self.process and self.process.is_running():
            rcs = reap_process_group(self.process.pid, self.log)
            self._rc = rcs.get(self.process.pid)

        self.process = None

        if self._rc is None:
            # Something else reaped it before we had a chance, so let's just "guess" at an error code.
            self._rc = -9



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import signal
import unittest
import time
import os
import shutil
from subprocess import Popen
from airflow.contrib.jobs.scheduler_client import EventSchedulerClient
from airflow.events.scheduler_events import StopSchedulerEvent
from airflow.utils.state import State
from notification_service.client import NotificationClient
from ai_flow.util.process_utils import get_all_children_pids
from ai_flow.endpoint.server.server_runner import AIFlowServerRunner
from ai_flow_plugins.scheduler_plugins.airflow.airflow_restful_util import AirFlowRestfulUtil
from ai_flow_plugins.tests import airflow_db_utils
from ai_flow_plugins.tests.airflow_scheduler_utils import start_airflow_scheduler_server, start_airflow_web_server
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server


project_path = os.path.dirname(__file__)


class TestAirFlowRestfulUtil(unittest.TestCase):
    master = None
    web_process = None
    sc_process = None

    @classmethod
    def setUpClass(cls) -> None:
        airflow_db_utils.clear_all()
        cls.ns_server = start_notification_server()
        config_file = project_path + '/master.yaml'
        cls.master = AIFlowServerRunner(config_file=config_file)
        cls.master.start()
        airflow_path = '/tmp/airflow_deploy'
        if os.path.exists(airflow_path):
            shutil.rmtree(airflow_path)
        os.makedirs(airflow_path)
        dag_dir = os.path.dirname(os.path.abspath(__file__))
        shutil.copyfile(dag_dir + '/dags/bash_1_dag.py', airflow_path + '/bash_1_dag.py')
        cls.sc_process = start_airflow_scheduler_server(file_path=airflow_path)
        cls.web_process: Popen = start_airflow_web_server()
        time.sleep(10)
        client = EventSchedulerClient(server_uri='localhost:50051', namespace='default')
        client.trigger_parse_dag('bash_1')
        for i in range(3):
            client.schedule_dag('bash_1', '')
            time.sleep(5)
        time.sleep(10)

    @classmethod
    def tearDownClass(cls) -> None:
        sub_pids = get_all_children_pids(cls.web_process.pid)
        try:
            for pid in sub_pids:
                os.kill(pid, signal.SIGKILL)
        except Exception:
            pass
        cls.web_process.kill()
        client = NotificationClient(server_uri="localhost:50051",
                                    default_namespace="default")
        client.send_event(StopSchedulerEvent(job_id=0).to_event())
        sub_pids = get_all_children_pids(cls.sc_process.pid)
        cls.sc_process.join()
        try:
            for pid in sub_pids:
                os.kill(pid, signal.SIGKILL)
        except Exception:
            pass
        cls.master.stop()
        stop_notification_server(cls.ns_server)

    def setUp(self):
        self.airflow_util = AirFlowRestfulUtil(endpoint_url='http://localhost:8080',
                                               user_name='admin',
                                               password='admin')

    def tearDown(self):
        self.master._clear_db()

    def test_list_dags(self):
        result = self.airflow_util.list_dags()
        self.assertEqual(1, len(result))

    def test_get_dag(self):
        result = self.airflow_util.get_dag('bash_1')
        self.assertTrue(result is not None)

    def test_pause_dag(self):
        result = self.airflow_util.set_dag_is_paused('bash_1', True)
        self.assertTrue(result.get('is_paused'))
        result = self.airflow_util.set_dag_is_paused('bash_1', False)
        self.assertFalse(result.get('is_paused'))

    def test_list_dagruns(self):
        result = self.airflow_util.list_dagruns('bash_1')
        self.assertEqual(3, len(result))

    def test_get_dagrun(self):
        result = self.airflow_util.list_dagruns('bash_1')
        result = self.airflow_util.get_dagrun(result[0].get('dag_id'), result[0].get('dag_run_id'))
        self.assertEqual(State.SUCCESS, result.get('state'))

    def test_list_taskInstances(self):
        result = self.airflow_util.list_dagruns('bash_1')
        result = self.airflow_util.list_task_instance(result[0].get('dag_id'), result[0].get('dag_run_id'))
        self.assertEqual(1, len(result))

    def test_get_taskInstances(self):
        result = self.airflow_util.list_dagruns('bash_1')
        result = self.airflow_util.get_task_instance(result[0].get('dag_id'),
                                                     result[0].get('dag_run_id'),
                                                     task_id='task_1')
        self.assertEqual(State.SUCCESS, result.get('state'))

    def test_list_taskExecutions(self):
        result = self.airflow_util.list_dagruns('bash_1')
        result = self.airflow_util.list_task_execution(result[0].get('dag_id'), result[0].get('dag_run_id'))
        self.assertEqual(1, len(result))

    def test_list_taskExecution_by_task_id(self):
        result = self.airflow_util.list_dagruns('bash_1')
        result = self.airflow_util.list_task_execution_by_task_id(result[0].get('dag_id'),
                                                                  result[0].get('dag_run_id'),
                                                                  'task_1')
        self.assertEqual(1, len(result))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import unittest

import grpc

from ai_flow.api import ai_flow_context
from ai_flow.api.ai_flow_context import init_ai_flow_context, init_notebook_context
from ai_flow.context.project_context import current_project_context, current_project_config
from ai_flow.context.workflow_config_loader import current_workflow_config
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server
from ai_flow.test.util.server_util import wait_for_server_started

_SQLITE_DB_FILE = 'aiflow.db'
_SQLITE_DB_URI = '%s%s' % ('sqlite:///', _SQLITE_DB_FILE)
_PORT = '50051'


class TestAIFlowContext(unittest.TestCase):
    def setUp(self):
        self.ns_server = start_notification_server()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        self.server = AIFlowServer(store_uri=_SQLITE_DB_URI, port=_PORT,
                                   start_meta_service=True,
                                   start_metric_service=False,
                                   start_model_center_service=False,
                                   start_scheduler_service=False)
        self.server.run()
        wait_for_server_started('localhost:{}'.format(_PORT))

    def tearDown(self):
        self.server.stop()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        stop_notification_server(self.ns_server)
        ai_flow_context.__init_notebook_context_flag__ = False
        ai_flow_context.__init_client_flag__ = False
        ai_flow_context.__init_context_flag__ = False

    def test_init_ai_flow_context(self):
        init_ai_flow_context()
        project_config = current_project_config()
        self.assertEqual('test_project', project_config.get_project_name())
        self.assertEqual('a', project_config.get('a'))
        project_context = current_project_context()
        self.assertEqual('test_project', project_context.project_name)
        workflow_config = current_workflow_config()
        self.assertEqual('test_ai_flow_context', workflow_config.workflow_name)
        self.assertEqual(5, len(workflow_config.job_configs))

        init_ai_flow_context()
        project_config = current_project_config()
        self.assertEqual('test_project', project_config.get_project_name())

        with self.assertRaises(Exception):
            init_notebook_context(project_config, workflow_config)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import os
from typing import List
import time
from ai_flow_plugins.job_plugins import python
from ai_flow_plugins.job_plugins.python.python_processor import ExecutionContext
from ai_flow import current_workflow_config


class PyProcessor1(python.PythonProcessor):

    def process(self, execution_context: ExecutionContext, input_list: List) -> List:
        print("Zhang san hello world!")
        word_count_file = os.path.dirname(__file__) + "/../../resources/word_count.txt"
        print(os.path.abspath(word_count_file))
        with open(word_count_file, 'r') as f:
            print(f.read())
        print(current_workflow_config().job_configs['task_1'].properties['a'])
        return []


class PyProcessor2(python.PythonProcessor):

    def process(self, execution_context: ExecutionContext, input_list: List) -> List:
        print("Li si hello world!")
        time.sleep(100)
        return []


#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Add task execution label

Revision ID: fb02e046d5e6
Revises: d47076ca1426
Create Date: 2021-09-02 11:45:12.937665

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = 'fb02e046d5e6'
down_revision = 'd47076ca1426'
branch_labels = None
depends_on = None


def upgrade():
    """Apply Add task execution label"""
    op.add_column('task_execution', sa.Column('execution_label', sa.Text, nullable=True))


def downgrade():
    """Unapply Add task execution label"""
    op.drop_column('task_execution', 'execution_label')



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import fcntl
import time
from typing import Text, Dict, Any
import oss2
from ai_flow.plugin_interface.blob_manager_interface import BlobManager


class OssBlobManager(BlobManager):
    """
    OssBlobManager is an implementation of BlobManager based on the oss file system
    OssBlobManager contains configuration items:
    1. access_key_id: The oss access key.
    2. access_key_secret: The oss access secret.
    3. bucket: The oss bucket name.
    4. root_directory: The upload directory of the project.
    """

    def __init__(self, config: Dict[str, Any]):
        super().__init__(config)
        if not self.root_dir:
            raise Exception('`root_directory` option of blob manager config is not configured.')
        self.ack_id = config.get('access_key_id', None)
        self.ack_secret = config.get('access_key_secret', None)
        self.endpoint = config.get('endpoint', None)
        self.bucket_name = config.get('bucket', None)
        self._bucket = None

    def upload(self, local_file_path: Text) -> Text:
        """
        Upload a given file to blob server. Uploaded file will be placed under self.root_dir.

        :param local_file_path: the path of file to be uploaded.
        :return the uri of the uploaded file in blob server.
        """
        file_name = os.path.basename(local_file_path)
        remote_file_path = os.path.join(self.root_dir, file_name)
        self.bucket.put_object_from_file(key=remote_file_path, filename=local_file_path)
        return self._get_absolute_uri(remote_file_path)

    def download(self, remote_file_path: Text, local_dir: Text) -> Text:
        """
        Download file from remote blob server to local directory.
        Only files located in self.root_dir can be downloaded by BlobManager.

        :param remote_file_path: The path of file to be downloaded.
        :param local_dir: the local directory.
        :return the local uri of the downloaded file.
        """
        self._check_remote_path_legality(remote_file_path)
        file_name = os.path.basename(remote_file_path)
        print(file_name)

        oss_object_key = remote_file_path.split("/", 1)[1]
        print(oss_object_key)

        local_file_path = os.path.join(local_dir, file_name)
        if not os.path.exists(local_file_path):
            self.log.debug("{} not exist".format(local_file_path))
            lock_file_path = os.path.join(local_dir, "{}.lock".format(file_name))
            lock_file = open(lock_file_path, 'w')
            self.log.debug("Locking file {}".format(lock_file_path))
            fcntl.flock(lock_file.fileno(), fcntl.LOCK_EX)
            self.log.debug("Locked file {}".format(lock_file_path))
            try:
                if not os.path.exists(local_file_path):
                    self.log.info("Downloading oss object: {}".format(oss_object_key))
                    self._get_oss_object(local_file_path, oss_object_key)
            except Exception as e:
                self.log.error("Failed to download oss file: {}".format(oss_object_key), exc_info=e)
            finally:
                fcntl.flock(lock_file.fileno(), fcntl.LOCK_UN)
                self.log.debug('Unlocked file {}'.format(lock_file_path))
                lock_file.close()
                if os.path.exists(lock_file_path):
                    try:
                        os.remove(lock_file_path)
                    except OSError as e:
                        self.log.warning("Failed to remove lock file: {}".format(lock_file_path), exc_info=e)
        else:
            self.log.info("Oss file: {} already exist at {}".format(oss_object_key, local_file_path))
        return local_file_path

    @property
    def bucket(self):
        if self._bucket:
            return self._bucket
        auth = oss2.Auth(self.ack_id, self.ack_secret)
        self._bucket = oss2.Bucket(auth, self.endpoint, self.bucket_name)
        return self._bucket

    def _get_oss_object(self, dest, oss_object_key, retry_sleep_sec=5):
        for i in range(3):
            try:
                self.bucket.get_object_to_file(oss_object_key, filename=dest)
                return
            except Exception as e:
                self.log.error("Downloading object {} failed, retrying {}/3 in {} second".format(
                    oss_object_key, i+1, retry_sleep_sec), exc_info=e)
                time.sleep(retry_sleep_sec)
        raise RuntimeError("Failed to download oss file: {}".format(oss_object_key))

    def _check_remote_path_legality(self, file_path: Text):
        """
        Check if the file can be downloaded by blob manager.

        :param file_path: The path of file to be checked.
        """
        absolute_root_dir = self._get_absolute_uri(self.root_dir)
        if not file_path.startswith(absolute_root_dir):
            raise Exception("Cannot download {} from blob server".format(file_path))

    def _get_absolute_uri(self, path):
        return '{}.{}/{}'.format(self.bucket_name, self.endpoint, path)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
import multiprocessing as mp
import threading
import time
from typing import Text, List

from setproctitle import setproctitle

from ai_flow.plugin_interface.scheduler_interface import SchedulerFactory
from ai_flow.scheduler_service.service.config import SchedulerServiceConfig
from ai_flow.scheduler_service.service.workflow_event_processor import Poison, WorkflowEventProcessor
from ai_flow.store.db.db_util import create_db_store
from notification_service.base_notification import EventWatcher, BaseEvent
from notification_service.client import NotificationClient

# Method to start the WorkflowEventProcessor
_MP_START_METHOD = 'spawn'


def _start_workflow_event_processor_process(conn, db_uri: Text, scheduler_service_config: SchedulerServiceConfig):
    setproctitle("WorkflowEventProcessor")
    store = create_db_store(db_uri)

    scheduler = SchedulerFactory.create_scheduler(scheduler_service_config.scheduler().scheduler_class(),
                                                  scheduler_service_config.scheduler().scheduler_config())
    processor = WorkflowEventProcessor(conn=conn, store=store, scheduler=scheduler)
    processor.run()


class WorkflowEventManager(object):
    """
    WorkflowEventManager
    """

    def __init__(self, notification_server_uri: Text,
                 db_uri: Text,
                 scheduler_service_config: SchedulerServiceConfig):
        self.db_uri = db_uri
        self.scheduler_service_config = scheduler_service_config
        self._notification_server_uri = notification_server_uri
        self._notification_client = None
        self.listen_event_handler = None

        self._stop = False
        self.event_processor_process = self._create_event_processor_process()
        self.process_watcher_thread = threading.Thread(target=self._watch_process)
        self.store = create_db_store(db_uri)

    @property
    def notification_client(self) -> NotificationClient:
        if self._notification_client is None:
            self._notification_client = NotificationClient(server_uri=self._notification_server_uri)
        return self._notification_client

    def _create_event_processor_process(self):
        # We use spawn to start the process to avoid problem of running grpc in multiple processes.
        # As we only spawn the new process once when the Scheduler service start, the performance drawback of spawn is
        # acceptable.
        ctx = mp.get_context(_MP_START_METHOD)
        self.processor_conn, self.conn = ctx.Pipe(False)
        return ctx.Process(target=_start_workflow_event_processor_process,
                           args=(self.processor_conn, self.db_uri, self.scheduler_service_config),
                           name="WorkflowEventProcessor")

    def start(self):
        logging.info("WorkflowEventManager start listening event")
        self._start_listen_events()
        self.event_processor_process.start()
        logging.info("Started WorkflowEventProcessor pid: {}".format(self.event_processor_process.pid))
        self.process_watcher_thread.start()

    def stop(self):
        logging.info("stopping WorkflowEventManager...")
        self._stop = True
        self.process_watcher_thread.join()
        self._stop_listen_events()
        if self.event_processor_process.is_alive():
            self.conn.send(Poison())
        self.event_processor_process.join()
        logging.info("WorkflowEventManager stopped...")

    def notify_event_arrived(self, event: BaseEvent):
        for i in range(10):
            try:
                self.conn.send(event)
                break
            except Exception as e:
                logging.warning("Error sending event to connection, retrying ({}/10)...".format(i),
                                exc_info=e)
                time.sleep(1)

    def _watch_process(self):
        while not self._stop:
            if self.event_processor_process.is_alive():
                time.sleep(1.)
                continue

            self.event_processor_process.join()
            logging.info("WorkflowEventProcessor pid: {} exited with code: {}".format(
                self.event_processor_process.pid,
                self.event_processor_process.exitcode))
            if self._stop:
                break
            logging.info("Restarting process")
            self._restart_event_listener()
            self.event_processor_process = self._create_event_processor_process()
            self.event_processor_process.start()
            logging.info("Process restarted with pid: {}".format(self.event_processor_process.pid))
        logging.info("Stop watching process")

    def _restart_event_listener(self):
        self._stop_listen_events()
        self._start_listen_events()

    def _stop_listen_events(self):
        logging.info("Stopping listening events")
        if self.listen_event_handler:
            self.listen_event_handler.stop()
            self.notification_client.stop_listen_events()

    def _start_listen_events(self):
        last_event_version = self._get_event_version_to_listen()
        logging.info("Starting listening events from version: {}".format(last_event_version))
        self.listen_event_handler = self.notification_client.start_listen_events(WorkflowEventWatcher(self),
                                                                                 version=last_event_version)

    def _get_event_version_to_listen(self):
        workflows = self.store.list_workflows()
        if workflows is None:
            return None
        workflow_event_versions = [workflow.last_event_version for workflow in workflows
                                   if workflow.last_event_version is not None]
        if len(workflow_event_versions) == 0:
            return None
        last_event_version = min(workflow_event_versions)
        return last_event_version


class WorkflowEventWatcher(EventWatcher):
    def __init__(self, workflow_event_manager: WorkflowEventManager):
        self._workflow_event_manager = workflow_event_manager

    def process(self, events: List[BaseEvent]):
        for event in events:
            self._workflow_event_manager.notify_event_arrived(event)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import os
import unittest

import ai_flow.settings
from ai_flow.settings import get_configuration, get_aiflow_home, get_configuration_file_path


class TestSettings(unittest.TestCase):

    def setUp(self) -> None:
        self.prev_AIFLOW_HOME = ai_flow.settings.AIFLOW_HOME

    def tearDown(self) -> None:
        ai_flow.settings.AIFLOW_HOME = self.prev_AIFLOW_HOME

    def test_get_configuration(self):
        ai_flow.settings.AIFLOW_HOME = os.path.dirname(__file__)
        config = get_configuration()
        self.assertEqual('sqlite:///aiflow.db', config.get_db_uri())
        self.assertEqual(50051, config.get_server_port())
        self.assertEqual('localhost:50052', config.get_notification_server_uri())

    def test_get_configuration_file_path(self):
        aiflow_home = os.path.dirname(__file__)
        ai_flow.settings.AIFLOW_HOME = aiflow_home
        self.assertEqual(os.path.join(aiflow_home, "aiflow_server.yaml"), get_configuration_file_path())

    def test_get_non_exist_configuration_file_path(self):
        ai_flow.settings.AIFLOW_HOME = '/non-exist-home'
        with self.assertRaises(FileNotFoundError):
            get_configuration_file_path()

    def test_get_aiflow_home(self):
        prev_home = os.environ['HOME']
        try:
            os.environ['HOME'] = '/home'
            self.assertEqual(os.path.join('/home', 'aiflow'), get_aiflow_home())
            os.environ['AIFLOW_HOME'] = '/aiflow_home'
            self.assertEqual('/aiflow_home', get_aiflow_home())

        finally:
            os.environ['HOME'] = prev_home
            if 'AIFLOW_HOME' in os.environ:
                del os.environ['AIFLOW_HOME']



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Increase length of pool name

Revision ID: b25a55525161
Revises: bbf4a7ad0465
Create Date: 2020-03-09 08:48:14.534700

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'b25a55525161'
down_revision = 'bbf4a7ad0465'
branch_labels = None
depends_on = None


def upgrade():
    """Increase column length of pool name from 50 to 256 characters"""
    # use batch_alter_table to support SQLite workaround
    with op.batch_alter_table('slot_pool', table_args=sa.UniqueConstraint('pool')) as batch_op:
        batch_op.alter_column('pool', type_=sa.String(256))


def downgrade():
    """Revert Increased length of pool name from 256 to 50 characters"""
    # use batch_alter_table to support SQLite workaround
    with op.batch_alter_table('slot_pool', table_args=sa.UniqueConstraint('pool')) as batch_op:
        batch_op.alter_column('pool', type_=sa.String(50))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from sqlalchemy import and_

from airflow.api_connexion import security
from airflow.api_connexion.exceptions import NotFound
from airflow.api_connexion.schemas.task_execution_schema import (
    TaskExecutionCollection,
    task_execution_collection_schema
)
from airflow.models import SlaMiss
from airflow.models.dagrun import DagRun as DR
from airflow.models.taskexecution import TaskExecution as TE
from airflow.security import permissions
from airflow.utils.session import provide_session


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_RUN),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_EXECUTION),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_INSTANCE),
    ]
)
@provide_session
def list_task_executions(dag_id: str, dag_run_id: str, session=None):
    """List task executions"""
    query = (
        session.query(TE)
        .filter(TE.dag_id == dag_id)
        .join(DR, and_(TE.dag_id == DR.dag_id, TE.execution_date == DR.execution_date))
        .filter(DR.run_id == dag_run_id)
        .outerjoin(
            SlaMiss,
            and_(
                SlaMiss.dag_id == TE.dag_id,
                SlaMiss.execution_date == TE.execution_date,
                SlaMiss.task_id == TE.task_id,
            ),
        )
        .add_entity(SlaMiss)
    )
    task_executions = query.all()
    if task_executions is None:
        raise NotFound("Task executions are not found for dag id({}) and dag run id({}).".format(dag_id, dag_run_id))
    total_count = len(task_executions)

    return task_execution_collection_schema.dump(
        TaskExecutionCollection(task_executions=task_executions, total_entries=total_count)
    )


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_RUN),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_EXECUTION),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_INSTANCE),
    ]
)
@provide_session
def list_task_executions_by_task_id(dag_id: str, dag_run_id: str, task_id: str, session=None):
    """List task executions"""
    query = (
        session.query(TE)
        .filter(TE.dag_id == dag_id)
        .join(DR, and_(TE.dag_id == DR.dag_id, TE.execution_date == DR.execution_date))
        .filter(DR.run_id == dag_run_id)
        .filter(TE.task_id == task_id)
        .outerjoin(
            SlaMiss,
            and_(
                SlaMiss.dag_id == TE.dag_id,
                SlaMiss.execution_date == TE.execution_date,
                SlaMiss.task_id == TE.task_id,
            ),
        )
        .add_entity(SlaMiss)
    )
    task_executions = query.all()
    if task_executions is None:
        raise NotFound("Task executions are not found for dag id({}), dag run id({}) and task id({}) ."
                       .format(dag_id, dag_run_id, task_id))
    total_count = len(task_executions)

    return task_execution_collection_schema.dump(
        TaskExecutionCollection(task_executions=task_executions, total_entries=total_count)
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import grpc
import mock
import os
import unittest

from ai_flow.api.context_extractor import WORKFLOW_EXECUTION_DEFAULT_CONTEXT
from ai_flow.workflow.control_edge import MeetAllEventCondition, MeetAnyEventCondition, WorkflowAction

from ai_flow.scheduler_service.service.config import SchedulerServiceConfig
from ai_flow.workflow.status import Status
from ai_flow.protobuf.message_pb2 import StateProto

from ai_flow.endpoint.client.scheduler_client import SchedulerClient
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.plugin_interface.scheduler_interface import JobExecutionInfo, WorkflowExecutionInfo, WorkflowInfo
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server
from ai_flow.test.test_util import MOCK_SCHEDULER_CLASS as SCHEDULER_CLASS

_SQLITE_DB_FILE = 'aiflow.db'
_SQLITE_DB_URI = '%s%s' % ('sqlite:///', _SQLITE_DB_FILE)
_PORT = '50051'


class TestSchedulerService(unittest.TestCase):
    def setUp(self):
        raw_config = {
            'scheduler': {
                'scheduler_class': SCHEDULER_CLASS,
            }
        }
        config = SchedulerServiceConfig(raw_config)
        self.ns_server = start_notification_server()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        self.server = AIFlowServer(store_uri=_SQLITE_DB_URI, port=_PORT,
                                   start_meta_service=False,
                                   start_metric_service=False,
                                   start_model_center_service=False,
                                   start_scheduler_service=True,
                                   scheduler_service_config=config)
        self.server.run()
        client = SchedulerClient("localhost:{}".format(_PORT))
        fut = grpc.channel_ready_future(client.channel)
        fut.result(2)

    def tearDown(self):
        self.server.stop()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        stop_notification_server(self.ns_server)

    def test_submit_workflow(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            instance.submit_workflow.return_value = WorkflowInfo(workflow_name='test_workflow')
            client = SchedulerClient("localhost:{}".format(_PORT))
            with self.assertRaises(Exception) as context:
                workflow = client.submit_workflow_to_scheduler(namespace='namespace', workflow_name='test_workflow',
                                                               workflow_json='')
            self.assertTrue('workflow json is empty' in str(context.exception))

    def test_pause_workflow(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.pause_workflow_scheduling.return_value = WorkflowInfo(workflow_name='test_workflow')
            client = SchedulerClient("localhost:{}".format(_PORT))
            workflow = client.pause_workflow_scheduling(namespace='namespace', workflow_name='test_workflow')
            self.assertTrue('test_workflow', workflow.name)

    def test_resume_workflow(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.resume_workflow_scheduling.return_value = WorkflowInfo(workflow_name='test_workflow')
            client = SchedulerClient("localhost:{}".format(_PORT))
            workflow = client.resume_workflow_scheduling(namespace='namespace', workflow_name='test_workflow')
            self.assertTrue('test_workflow', workflow.name)

    def test_start_new_workflow_execution(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.start_new_workflow_execution.return_value \
                = WorkflowExecutionInfo(workflow_execution_id='id', status=Status.INIT)
            client = SchedulerClient("localhost:{}".format(_PORT))
            workflow_execution = client.start_new_workflow_execution(namespace='namespace',
                                                                     workflow_name='test_workflow')
            args, kwargs = instance.start_new_workflow_execution.call_args
            self.assertEqual(('namespace', 'test_workflow', None), args)
            self.assertEqual('id', workflow_execution.execution_id)
            self.assertEqual(StateProto.INIT, workflow_execution.execution_state)
            self.assertFalse(workflow_execution.HasField('context'))

    def test_start_new_workflow_execution_with_context(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            self.server.scheduler_service._scheduler = mockScheduler

            mockScheduler.start_new_workflow_execution.return_value \
                = WorkflowExecutionInfo(workflow_execution_id='id', status=Status.INIT, context='test_context')
            client = SchedulerClient("localhost:{}".format(_PORT))
            workflow_execution = client.start_new_workflow_execution(namespace='namespace',
                                                                     workflow_name='test_workflow',
                                                                     context='test_context')
            args, kwargs = mockScheduler.start_new_workflow_execution.call_args
            self.assertEqual(('namespace', 'test_workflow', 'test_context'), args)
            self.assertEqual('id', workflow_execution.execution_id)
            self.assertEqual(StateProto.INIT, workflow_execution.execution_state)
            self.assertEqual('test_context', workflow_execution.context.value)

    def test_start_new_workflow_execution_on_event(self):
        store = self.server.scheduler_service.store
        project_meta = store.register_project('namespace', 'example.com')
        store.register_workflow('test_workflow', project_meta.uuid)

        client = SchedulerClient("localhost:{}".format(_PORT))
        c1 = MeetAllEventCondition().add_event('k1', 'v1')
        c2 = MeetAnyEventCondition().add_event('k2', 'v2')
        workflow_proto = client.start_new_workflow_execution_on_events(namespace='namespace',
                                                                       workflow_name='test_workflow',
                                                                       event_conditions=[c1, c2])
        self.assertEqual('namespace', workflow_proto.namespace)
        self.assertEqual('test_workflow', workflow_proto.name)

        workflow_meta = store.get_workflow_by_name('namespace', 'test_workflow')
        start_event_condition = [rule.event_condition for rule in workflow_meta.scheduling_rules if
                                 rule.action == WorkflowAction.START]
        self.assertIn(c1, start_event_condition)
        self.assertIn(c2, start_event_condition)

        client.start_new_workflow_execution_on_events(namespace='namespace',
                                                      workflow_name='test_workflow',
                                                      event_conditions=[])
        workflow_meta = store.get_workflow_by_name('namespace', 'test_workflow')
        self.assertEqual(0, len(workflow_meta.scheduling_rules))
        self.assertEqual([], workflow_meta.get_condition(WorkflowAction.START))

        # default context state is registered
        context_state = store.get_workflow_context_event_handler_state('namespace', 'test_workflow',
                                                                       WORKFLOW_EXECUTION_DEFAULT_CONTEXT)
        self.assertIsNotNone(context_state)
        self.assertEqual(WORKFLOW_EXECUTION_DEFAULT_CONTEXT, context_state.context)

    def test_stop_workflow_execution_on_event(self):
        store = self.server.scheduler_service.store
        project_meta = store.register_project('namespace', 'example.com')
        store.register_workflow('test_workflow', project_meta.uuid)

        client = SchedulerClient("localhost:{}".format(_PORT))
        c1 = MeetAllEventCondition().add_event('k1', 'v1')
        c2 = MeetAnyEventCondition().add_event('k2', 'v2')
        workflow_proto = client.stop_workflow_execution_on_events(namespace='namespace',
                                                                  workflow_name='test_workflow',
                                                                  event_conditions=[c1, c2])
        self.assertEqual('namespace', workflow_proto.namespace)
        self.assertEqual('test_workflow', workflow_proto.name)

        workflow_meta = store.get_workflow_by_name('namespace', 'test_workflow')
        start_event_condition = [rule.event_condition for rule in workflow_meta.scheduling_rules if
                                 rule.action == WorkflowAction.STOP]
        self.assertIn(c1, start_event_condition)
        self.assertIn(c2, start_event_condition)

        client.stop_workflow_execution_on_events(namespace='namespace',
                                                 workflow_name='test_workflow',
                                                 event_conditions=[])
        workflow_meta = store.get_workflow_by_name('namespace', 'test_workflow')
        self.assertEqual(0, len(workflow_meta.scheduling_rules))
        self.assertEqual([], workflow_meta.get_condition(WorkflowAction.STOP))

        # default context state is registered
        context_state = store.get_workflow_context_event_handler_state('namespace', 'test_workflow',
                                                                       WORKFLOW_EXECUTION_DEFAULT_CONTEXT)
        self.assertIsNotNone(context_state)
        self.assertEqual(WORKFLOW_EXECUTION_DEFAULT_CONTEXT, context_state.context)

    def test_kill_all_workflow_execution(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.stop_all_workflow_execution.return_value \
                = [WorkflowExecutionInfo(workflow_execution_id='id_1', status=Status.INIT),
                   WorkflowExecutionInfo(workflow_execution_id='id_2', status=Status.INIT)]
            client = SchedulerClient("localhost:{}".format(_PORT))
            workflow_execution_list = client.kill_all_workflow_executions(namespace='namespace',
                                                                          workflow_name='test_workflow')
            self.assertEqual(2, len(workflow_execution_list))

    def test_kill_workflow_execution(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.stop_workflow_execution.return_value \
                = WorkflowExecutionInfo(workflow_execution_id='id', status=Status.RUNNING)
            client = SchedulerClient("localhost:{}".format(_PORT))
            workflow_execution = client.kill_workflow_execution(execution_id='id')
            self.assertEqual('id', workflow_execution.execution_id)
            self.assertEqual(StateProto.RUNNING, workflow_execution.execution_state)

    def test_get_workflow_execution(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.get_workflow_execution.return_value \
                = WorkflowExecutionInfo(workflow_execution_id='id', status=Status.INIT)
            client = SchedulerClient("localhost:{}".format(_PORT))
            workflow_execution = client.get_workflow_execution(execution_id='id')
            self.assertEqual('id', workflow_execution.execution_id)
            self.assertEqual(StateProto.INIT, workflow_execution.execution_state)

    def test_list_workflow_executions(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.list_workflow_executions.return_value \
                = [WorkflowExecutionInfo(workflow_execution_id='id_1', status=Status.INIT),
                   WorkflowExecutionInfo(workflow_execution_id='id_2', status=Status.INIT)]
            client = SchedulerClient("localhost:{}".format(_PORT))
            workflow_execution_list = client.list_workflow_executions(namespace='namespace',
                                                                      workflow_name='test_workflow')
            self.assertEqual(2, len(workflow_execution_list))

    def test_start_job(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.start_job_execution.return_value \
                = JobExecutionInfo(job_name='job_name',
                                   status=Status.RUNNING,
                                   workflow_execution=WorkflowExecutionInfo(workflow_execution_id='id',
                                                                            status=Status.INIT))
            client = SchedulerClient("localhost:{}".format(_PORT))
            job = client.start_job(job_name='job_name', execution_id='id')
            self.assertEqual('job_name', job.name)
            self.assertEqual(StateProto.RUNNING, job.job_state)
            self.assertEqual('id', job.workflow_execution.execution_id)
            self.assertEqual(StateProto.INIT, job.workflow_execution.execution_state)

    def test_stop_job(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.stop_job_execution.return_value \
                = JobExecutionInfo(job_name='job_name',
                                   status=Status.RUNNING,
                                   workflow_execution=WorkflowExecutionInfo(workflow_execution_id='id',
                                                                            status=Status.INIT))
            client = SchedulerClient("localhost:{}".format(_PORT))
            job = client.stop_job(job_name='job_name', execution_id='id')
            self.assertEqual('job_name', job.name)
            self.assertEqual(StateProto.RUNNING, job.job_state)
            self.assertEqual('id', job.workflow_execution.execution_id)
            self.assertEqual(StateProto.INIT, job.workflow_execution.execution_state)

    def test_restart_job(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.restart_job_execution.return_value \
                = JobExecutionInfo(job_name='job_name',
                                   status=Status.RUNNING,
                                   workflow_execution=WorkflowExecutionInfo(workflow_execution_id='id',
                                                                            status=Status.INIT))
            client = SchedulerClient("localhost:{}".format(_PORT))
            job = client.restart_job(job_name='job_name', execution_id='id')
            self.assertEqual('job_name', job.name)
            self.assertEqual(StateProto.RUNNING, job.job_state)
            self.assertEqual('id', job.workflow_execution.execution_id)
            self.assertEqual(StateProto.INIT, job.workflow_execution.execution_state)

    def test_get_job(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.get_job_executions.return_value \
                = [JobExecutionInfo(job_name='job_name',
                                    status=Status.RUNNING,
                                    workflow_execution=WorkflowExecutionInfo(workflow_execution_id='id',
                                                                             status=Status.INIT))]
            client = SchedulerClient("localhost:{}".format(_PORT))
            jobs = client.get_job(job_name='job_name', execution_id='id')
            self.assertEqual('job_name', jobs[0].name)
            self.assertEqual(StateProto.RUNNING, jobs[0].job_state)
            self.assertEqual('id', jobs[0].workflow_execution.execution_id)
            self.assertEqual(StateProto.INIT, jobs[0].workflow_execution.execution_state)

    def test_list_jobs(self):
        with mock.patch(SCHEDULER_CLASS) as mockScheduler:
            instance = mockScheduler.return_value
            self.server.scheduler_service._scheduler = instance

            instance.list_job_executions.return_value \
                = [JobExecutionInfo(job_name='job_name_1',
                                    status=Status.RUNNING,
                                    workflow_execution=WorkflowExecutionInfo(workflow_execution_id='id',
                                                                             status=Status.INIT)),
                   JobExecutionInfo(job_name='job_name_2',
                                    status=Status.RUNNING,
                                    workflow_execution=WorkflowExecutionInfo(workflow_execution_id='id',
                                                                             status=Status.INIT))]
            client = SchedulerClient("localhost:{}".format(_PORT))
            job_list = client.list_jobs(execution_id='id')
            self.assertEqual(2, len(job_list))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import io
import logging
import os
import unittest
from contextlib import redirect_stdout

from ai_flow.cli import cli_parser
from ai_flow.cli.commands.config_command import config_init, config_get_value, config_list
from ai_flow.settings import AIFLOW_HOME

logger = logging.getLogger(__name__)


class TestCliConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.parser = cli_parser.get_parser()

    def setUp(self) -> None:
        self.config_file = AIFLOW_HOME + '/aiflow_server.yaml'
        if os.path.exists(self.config_file):
            os.remove(self.config_file)

    def test_cli_config_get_value(self):
        config_init(self.parser.parse_args(['config', 'init']))
        with redirect_stdout(io.StringIO()) as stdout:
            config_get_value(self.parser.parse_args(['config', 'get-value', 'server_port']))
            self.assertEquals('50051', str.splitlines(stdout.getvalue())[0])
        with self.assertRaises(SystemExit):
            config_get_value(self.parser.parse_args(['config', 'get-value', 'server_ip']))

    def test_cli_config_init(self):
        with self.assertLogs('ai_flow.cli.commands.config_command', 'INFO') as log:
            config_init(self.parser.parse_args(['config', 'init']))
            log_output = '\n'.join(log.output)
            self.assertIn('AIFlow server config generated at {}.'.format(self.config_file), log_output)

            config_init(self.parser.parse_args(['config', 'init']))
            log_output = "\n".join(log.output)
            self.assertIn('AIFlow server config has already initialized at {}.'.format(self.config_file),
                          log_output)

    def test_cli_config_list(self):
        config_init(self.parser.parse_args(['config', 'init']))
        with redirect_stdout(io.StringIO()) as stdout:
            config_list(self.parser.parse_args(['config', 'list']))
            self.assertIn('# port of AIFlow server\nserver_port: 50051\n', stdout.getvalue())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import List, Text

from pyflink.table.statement_set import StatementSet
from pyflink.table.table_environment import TableEnvironment
from pyflink.table.table_result import TableResult
from pyflink.version import __version__

from ai_flow_plugins.job_plugins.flink.flink_version import INCLUDE_DATASET_VERSIONS

if __version__ in INCLUDE_DATASET_VERSIONS:
    from pyflink.table.table_environment import BatchTableEnvironment, StreamTableEnvironment


class WrappedTableEnvironmentContext:
    """
    WrappedTableEnvironmentContext provides container to store extra information for WrappedTableEnvironment,
    including list of TableResult for each execution.
    It can help to provide these functions:
    - Lists the ids of the Flink jobs.
    - Waits for the execution result of the Flink job.
    """

    def __init__(self):
        self.execute_sql_results: List[TableResult] = []

    def add_execute_sql_result(self, result: TableResult):
        self.execute_sql_results.append(result)

    def get_job_ids(self) -> List[Text]:
        result = []
        for r in self.execute_sql_results:
            job_client = r.get_job_client()
            if job_client is not None:
                job_id = str(job_client.get_job_id())
                result.append(job_id)
        return result

    def wait_execution_results(self):
        for r in self.execute_sql_results:
            job_client = r.get_job_client()
            if job_client is not None:
                job_client.get_job_execution_result().result()


class WrappedTableEnvironment(TableEnvironment):
    """
    WrappedTableEnvironment overrides the following functions to collect information needed for 'flink_run_main.py':
    - execute_sql(self, stmt: str) -> TableResult
    - create_statement_set(self) -> 'WrappedStatementSet':
    """
    wrapped_context = WrappedTableEnvironmentContext()

    def execute_sql(self, stmt: str) -> TableResult:
        result = super().execute_sql(stmt)
        self.wrapped_context.add_execute_sql_result(result)
        return result

    def create_statement_set(self) -> 'WrappedStatementSet':
        _s_set = super().create_statement_set()
        s_set = WrappedStatementSet(_s_set._j_statement_set, _s_set._t_env)
        return s_set

    def wait_execution_results(self):
        self.wrapped_context.wait_execution_results()

    if __version__ not in INCLUDE_DATASET_VERSIONS:
        @staticmethod
        def create_from(t_env: TableEnvironment) -> 'WrappedTableEnvironment':
            return WrappedTableEnvironment(j_tenv=t_env._j_tenv)


if __version__ in INCLUDE_DATASET_VERSIONS:
    class WrappedBatchTableEnvironment(BatchTableEnvironment, WrappedTableEnvironment):

        def __init__(self, j_tenv):
            super().__init__(j_tenv)

        @staticmethod
        def create_from(t_env: BatchTableEnvironment) -> 'WrappedBatchTableEnvironment':
            return WrappedBatchTableEnvironment(t_env._j_tenv)


    class WrappedStreamTableEnvironment(StreamTableEnvironment, WrappedTableEnvironment):

        def __init__(self, j_tenv):
            super().__init__(j_tenv)

        @staticmethod
        def create_from(t_env: StreamTableEnvironment) -> 'WrappedStreamTableEnvironment':
            return WrappedStreamTableEnvironment(t_env._j_tenv)


class WrappedStatementSetContext:
    """
    WrappedStatementSetContext provides container to store extra information for WrappedStatementSet,
    including list of TableResult for each execution.
    It can help to provide these functions:
    - Lists the ids of the Flink jobs.
    - Waits for the execution result of the Flink job.
    - Decides whether 'StatementSet' needs to call 'execute()' method based on whether it contains 'execute_sql'.
    """

    def __init__(self):
        self.execute_results: List[TableResult] = []
        self.need_execute = False

    def add_execute_result(self, result: TableResult):
        self.execute_results.append(result)

    def get_job_ids(self) -> List[Text]:
        result = []
        for r in self.execute_results:
            job_client = r.get_job_client()
            if job_client is not None:
                job_id = str(job_client.get_job_id())
                result.append(job_id)
        return result

    def wait_execution_results(self):
        for r in self.execute_results:
            job_client = r.get_job_client()
            if job_client is not None:
                job_client.get_job_execution_result().result()


class WrappedStatementSet(StatementSet):
    """
    WrappedStatementSet overrides the following functions to collect information needed for 'flink_run_main.py':
    - execute(self) -> TableResult
    - add_insert_sql(self, stmt: str) -> 'StatementSet':
    """

    wrapped_context = WrappedStatementSetContext()

    def execute(self) -> TableResult:
        result = super().execute()
        self.wrapped_context.add_execute_result(result)
        self.wrapped_context.need_execute = False
        return result

    def add_insert_sql(self, stmt: str) -> 'StatementSet':
        result = super().add_insert_sql(stmt)
        self.wrapped_context.need_execute = True
        return result

    def add_insert(self, target_path, table, overwrite: bool = False) -> 'StatementSet':
        result = super().add_insert(target_path, table, overwrite=overwrite)
        self.wrapped_context.need_execute = True
        return result

    def wait_execution_results(self):
        self.wrapped_context.wait_execution_results()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import json
import os
import unittest
from unittest import mock

from cloudpickle import cloudpickle

from ai_flow import WorkflowMeta
from ai_flow.frontend.web_server import generate_graph, AIFlowWebServerConfig, version
from ai_flow.test.scheduler_service.service.test_workflow_event_processor import MyContextExtractor
from ai_flow.version import __version__
from ai_flow.workflow.control_edge import WorkflowSchedulingRule, WorkflowAction, MeetAllEventCondition


class TestWebServer(unittest.TestCase):

    def test_version(self):
        self.assertEqual(version(), __version__)

    def test_generate_acyclic_graph(self):
        context_extractor = MyContextExtractor()
        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.STOP)
        workflow_meta = WorkflowMeta('workflow', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                                     scheduling_rules=[rule],
                                     graph='{"__af_object_type__": "jsonable", "__class__": "AIGraph", "__module__": "ai_flow.ai_graph.ai_graph", "_context_extractor": {"__af_object_type__": "jsonable", "__class__": "DailyWorkflowContextExtractor", "__module__": "__main__"}, "edges": {"AINode_0": [{"__af_object_type__": "jsonable", "__class__": "DataEdge", "__module__": "ai_flow.ai_graph.data_edge", "destination": "AINode_0", "port": 0, "source": "ReadDatasetNode_0"}], "AINode_1": [{"__af_object_type__": "jsonable", "__class__": "DataEdge", "__module__": "ai_flow.ai_graph.data_edge", "destination": "AINode_1", "port": 0, "source": "ReadDatasetNode_1"}], "AINode_2": [{"__af_object_type__": "jsonable", "__class__": "DataEdge", "__module__": "ai_flow.ai_graph.data_edge", "destination": "AINode_2", "port": 0, "source": "AINode_1"}], "WriteDatasetNode_1": [{"__af_object_type__": "jsonable", "__class__": "DataEdge", "__module__": "ai_flow.ai_graph.data_edge", "destination": "WriteDatasetNode_1", "port": 0, "source": "AINode_2"}], "daily_validate": [{"__af_object_type__": "jsonable", "__class__": "ControlEdge", "__module__": "ai_flow.workflow.control_edge", "destination": "daily_validate", "scheduling_rule": {"__af_object_type__": "jsonable", "__class__": "JobSchedulingRule", "__module__": "ai_flow.workflow.control_edge", "action": "START", "event_condition": {"__af_object_type__": "jsonable", "__class__": "MeetAnyEventCondition", "__module__": "ai_flow.workflow.control_edge", "condition_type": "MEET_ANY", "events": [{"__af_object_type__": "jsonable", "__class__": "EventMeetConfig", "__module__": "ai_flow.workflow.control_edge", "event_key": "daily_workflow.daily_training", "event_type": "JOB_STATUS_CHANGED", "event_value": "FINISHED", "life": "ONCE", "namespace": "workflow_on_event", "sender": "daily_training", "value_condition": "EQUALS"}]}}, "source": "*"}]}, "name": null, "node_id": "AIGraph_0", "nodes": {"AINode_0": {"__af_object_type__": "jsonable", "__class__": "AINode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "daily_training", "job_type": "python", "properties": {"entry_module_path": "daily_workflow"}}, "name": null, "node_config": {"base_model_info": null, "model_info": null, "name": null, "node_type": "train", "properties": null}, "node_id": "AINode_0", "output_num": 0, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003c__main__\nDailyTrainingTrain\nq\u0000)\u0081q\u0001.", "__module__": "builtins"}, "properties": {}}, "AINode_1": {"__af_object_type__": "jsonable", "__class__": "AINode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "daily_validate", "job_type": "python", "properties": {"entry_module_path": "daily_workflow"}}, "name": null, "node_config": {"name": null, "node_type": "transform", "properties": null}, "node_id": "AINode_1", "output_num": 1, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003c__main__\nDailyValidateTransformer\nq\u0000)\u0081q\u0001.", "__module__": "builtins"}, "properties": {}}, "AINode_2": {"__af_object_type__": "jsonable", "__class__": "AINode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "daily_validate", "job_type": "python", "properties": {"entry_module_path": "daily_workflow"}}, "name": null, "node_config": {"name": null, "node_type": "user_define_operation", "properties": null}, "node_id": "AINode_2", "output_num": 1, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003c__main__\nDailyValidate\nq\u0000)\u0081q\u0001}q\u0002(X\n\u0000\u0000\u0000model_nameq\u0003NX\n\u0000\u0000\u0000model_pathq\u0004NX\r\u0000\u0000\u0000model_versionq\u0005Nub.", "__module__": "builtins"}, "properties": {}}, "ReadDatasetNode_0": {"__af_object_type__": "jsonable", "__class__": "ReadDatasetNode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "daily_training", "job_type": "python", "properties": {"entry_module_path": "daily_workflow"}}, "name": null, "node_config": {"dataset": {"__af_object_type__": "jsonable", "__class__": "DatasetMeta", "__module__": "ai_flow.meta.dataset_meta", "catalog_connection_uri": null, "catalog_database": null, "catalog_name": null, "catalog_table": null, "catalog_type": null, "create_time": 1631259398354, "data_format": null, "description": null, "name": "daily_data", "properties": null, "schema": {"__af_object_type__": "jsonable", "__class__": "Schema", "__module__": "ai_flow.meta.dataset_meta", "name_list": null, "type_list": null}, "update_time": 1631259398354, "uri": "/tmp/daily_data", "uuid": 6}, "name": null, "node_type": "read_dataset", "properties": null}, "node_id": "ReadDatasetNode_0", "output_num": 1, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003c__main__\nDailyTrainingReader\nq\u0000)\u0081q\u0001.", "__module__": "builtins"}, "properties": {}}, "ReadDatasetNode_1": {"__af_object_type__": "jsonable", "__class__": "ReadDatasetNode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "daily_validate", "job_type": "python", "properties": {"entry_module_path": "daily_workflow"}}, "name": null, "node_config": {"dataset": {"__af_object_type__": "jsonable", "__class__": "DatasetMeta", "__module__": "ai_flow.meta.dataset_meta", "catalog_connection_uri": null, "catalog_database": null, "catalog_name": null, "catalog_table": null, "catalog_type": null, "create_time": 1631259398087, "data_format": null, "description": null, "name": "mnist_evaluate", "properties": null, "schema": {"__af_object_type__": "jsonable", "__class__": "Schema", "__module__": "ai_flow.meta.dataset_meta", "name_list": null, "type_list": null}, "update_time": 1631259398087, "uri": "/Users/sxnan/workspace/flinkml/flink-ai-extended/flink-ai-flow/examples/dataset_data/mnist_evaluate.npz", "uuid": 2}, "name": null, "node_type": "read_dataset", "properties": null}, "node_id": "ReadDatasetNode_1", "output_num": 1, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003c__main__\nDailyValidateReader\nq\u0000)\u0081q\u0001.", "__module__": "builtins"}, "properties": {}}, "WriteDatasetNode_0": {"__af_object_type__": "jsonable", "__class__": "WriteDatasetNode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "daily_training", "job_type": "python", "properties": {"entry_module_path": "daily_workflow"}}, "name": null, "node_config": {"dataset": {"__af_object_type__": "jsonable", "__class__": "DatasetMeta", "__module__": "ai_flow.meta.dataset_meta", "catalog_connection_uri": null, "catalog_database": null, "catalog_name": null, "catalog_table": null, "catalog_type": null, "create_time": 1631504656736, "data_format": null, "description": null, "name": "daily_train_result", "properties": null, "schema": {"__af_object_type__": "jsonable", "__class__": "Schema", "__module__": "ai_flow.meta.dataset_meta", "name_list": null, "type_list": null}, "update_time": 1631504656736, "uri": null, "uuid": 8}, "name": null, "node_type": "write_dataset", "properties": null}, "node_id": "WriteDatasetNode_0", "output_num": 0, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003c__main__\nDummyWriter\nq\u0000)\u0081q\u0001.", "__module__": "builtins"}, "properties": {}}, "WriteDatasetNode_1": {"__af_object_type__": "jsonable", "__class__": "WriteDatasetNode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "daily_validate", "job_type": "python", "properties": {"entry_module_path": "daily_workflow"}}, "name": null, "node_config": {"dataset": {"__af_object_type__": "jsonable", "__class__": "DatasetMeta", "__module__": "ai_flow.meta.dataset_meta", "catalog_connection_uri": null, "catalog_database": null, "catalog_name": null, "catalog_table": null, "catalog_type": null, "create_time": 1631504656808, "data_format": null, "description": null, "name": "daily_validate_result", "properties": null, "schema": {"__af_object_type__": "jsonable", "__class__": "Schema", "__module__": "ai_flow.meta.dataset_meta", "name_list": null, "type_list": null}, "update_time": 1631504656808, "uri": null, "uuid": 9}, "name": null, "node_type": "write_dataset", "properties": null}, "node_id": "WriteDatasetNode_1", "output_num": 0, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003c__main__\nDummyWriter\nq\u0000)\u0081q\u0001.", "__module__": "builtins"}, "properties": {}}}, "output_num": 0, "properties": {}}')
        workflow_graph = generate_graph(workflow_meta)
        self.assertIsNotNone(workflow_graph)
        graph_nodes = json.loads(workflow_graph)
        for graph_node in graph_nodes:
            if graph_node['id'] == 'daily_data':
                self.assertEqual(graph_node['layer'], 1)
            if graph_node['id'] == 'daily_train_result':
                self.assertEqual(graph_node['layer'], 1)
            if graph_node['id'] == 'daily_training':
                self.assertEqual(graph_node['layer'], 2)
            if graph_node['id'] == 'mnist_evaluate':
                self.assertEqual(graph_node['layer'], 2)
            if graph_node['id'] == 'daily_validate':
                self.assertEqual(graph_node['layer'], 3)
            if graph_node['id'] == 'daily_validate_result':
                self.assertEqual(graph_node['layer'], 4)

    def test_generate_ring_graph(self):
        context_extractor = MyContextExtractor()
        rule = WorkflowSchedulingRule(MeetAllEventCondition().add_event('k', 'v', namespace='test_namespace'),
                                      WorkflowAction.STOP)
        workflow_meta = WorkflowMeta('workflow', 0, context_extractor_in_bytes=cloudpickle.dumps(context_extractor),
                                     scheduling_rules=[rule],
                                     graph='{"__af_object_type__": "jsonable", "__class__": "AIGraph", "__module__": "ai_flow.ai_graph.ai_graph", "_context_extractor": {"__af_object_type__": "jsonable", "__class__": "BroadcastAllContextExtractor", "__module__": "ai_flow.api.context_extractor"}, "edges": {"task_2": [{"__af_object_type__": "jsonable", "__class__": "ControlEdge", "__module__": "ai_flow.workflow.control_edge", "destination": "task_2", "scheduling_rule": {"__af_object_type__": "jsonable", "__class__": "JobSchedulingRule", "__module__": "ai_flow.workflow.control_edge", "action": "START", "event_condition": {"__af_object_type__": "jsonable", "__class__": "MeetAnyEventCondition", "__module__": "ai_flow.workflow.control_edge", "condition_type": "MEET_ANY", "events": [{"__af_object_type__": "jsonable", "__class__": "EventMeetConfig", "__module__": "ai_flow.workflow.control_edge", "event_key": "simple_workflow.task_1", "event_type": "JOB_STATUS_CHANGED", "event_value": "FINISHED", "life": "ONCE", "namespace": "celery_examples", "sender": "task_1", "value_condition": "EQUALS"}]}}, "source": "*"}, {"__af_object_type__": "jsonable", "__class__": "ControlEdge", "__module__": "ai_flow.workflow.control_edge", "destination": "task_2", "scheduling_rule": {"__af_object_type__": "jsonable", "__class__": "JobSchedulingRule", "__module__": "ai_flow.workflow.control_edge", "action": "STOP", "event_condition": {"__af_object_type__": "jsonable", "__class__": "MeetAnyEventCondition", "__module__": "ai_flow.workflow.control_edge", "condition_type": "MEET_ANY", "events": [{"__af_object_type__": "jsonable", "__class__": "EventMeetConfig", "__module__": "ai_flow.workflow.control_edge", "event_key": "simple_workflow.task_3", "event_type": "JOB_STATUS_CHANGED", "event_value": "FINISHED", "life": "ONCE", "namespace": "celery_examples", "sender": "task_3", "value_condition": "EQUALS"}]}}, "source": "*"}], "task_3": [{"__af_object_type__": "jsonable", "__class__": "ControlEdge", "__module__": "ai_flow.workflow.control_edge", "destination": "task_3", "scheduling_rule": {"__af_object_type__": "jsonable", "__class__": "JobSchedulingRule", "__module__": "ai_flow.workflow.control_edge", "action": "START", "event_condition": {"__af_object_type__": "jsonable", "__class__": "MeetAnyEventCondition", "__module__": "ai_flow.workflow.control_edge", "condition_type": "MEET_ANY", "events": [{"__af_object_type__": "jsonable", "__class__": "EventMeetConfig", "__module__": "ai_flow.workflow.control_edge", "event_key": "simple_workflow.task_2", "event_type": "JOB_STATUS_CHANGED", "event_value": "RUNNING", "life": "ONCE", "namespace": "celery_examples", "sender": "task_2", "value_condition": "EQUALS"}]}}, "source": "*"}]}, "name": null, "node_id": "AIGraph_0", "nodes": {"AINode_0": {"__af_object_type__": "jsonable", "__class__": "AINode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "task_1", "job_type": "bash", "properties": {"entry_module_path": "simple_workflow"}}, "name": null, "node_config": {"name": null, "node_type": "user_define_operation", "properties": null}, "node_id": "AINode_0", "output_num": 1, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003cai_flow_plugins.job_plugins.bash.bash_processor\nBashProcessor\nq\u0000)\u0081q\u0001}q\u0002(X\f\u0000\u0000\u0000bash_commandq\u0003X\u0011\u0000\u0000\u0000echo before_sleepq\u0004X\u000f\u0000\u0000\u0000output_encodingq\u0005X\u0005\u0000\u0000\u0000utf-8q\u0006ub.", "__module__": "builtins"}, "properties": {}}, "AINode_1": {"__af_object_type__": "jsonable", "__class__": "AINode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "task_2", "job_type": "bash", "properties": {"entry_module_path": "simple_workflow"}}, "name": null, "node_config": {"name": null, "node_type": "user_define_operation", "properties": null}, "node_id": "AINode_1", "output_num": 1, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003cai_flow_plugins.job_plugins.bash.bash_processor\nBashProcessor\nq\u0000)\u0081q\u0001}q\u0002(X\f\u0000\u0000\u0000bash_commandq\u0003X\t\u0000\u0000\u0000sleep 100q\u0004X\u000f\u0000\u0000\u0000output_encodingq\u0005X\u0005\u0000\u0000\u0000utf-8q\u0006ub.", "__module__": "builtins"}, "properties": {}}, "AINode_2": {"__af_object_type__": "jsonable", "__class__": "AINode", "__module__": "ai_flow.ai_graph.ai_node", "config": {"__af_object_type__": "jsonable", "__class__": "JobConfig", "__module__": "ai_flow.workflow.job_config", "job_label_report_interval": 5.0, "job_name": "task_3", "job_type": "bash", "properties": {"entry_module_path": "simple_workflow"}}, "name": null, "node_config": {"name": null, "node_type": "user_define_operation", "properties": null}, "node_id": "AINode_2", "output_num": 1, "processor": {"__af_object_type__": "bytes", "__class__": "bytes", "__data__": "\u0080\u0003cai_flow_plugins.job_plugins.bash.bash_processor\nBashProcessor\nq\u0000)\u0081q\u0001}q\u0002(X\f\u0000\u0000\u0000bash_commandq\u0003X\b\u0000\u0000\u0000sleep 10q\u0004X\u000f\u0000\u0000\u0000output_encodingq\u0005X\u0005\u0000\u0000\u0000utf-8q\u0006ub.", "__module__": "builtins"}, "properties": {}}}, "output_num": 0, "properties": {}}')
        workflow_graph = generate_graph(workflow_meta)
        self.assertIsNotNone(workflow_graph)
        graph_nodes = json.loads(workflow_graph)
        for graph_node in graph_nodes:
            if graph_node['id'] == 'task_1':
                self.assertEqual(graph_node['layer'], 1)
            if graph_node['id'] == 'task_2':
                self.assertEqual(graph_node['layer'], 2)
            if graph_node['id'] == 'task_3':
                self.assertEqual(graph_node['layer'], 1)

    def test_web_server_read_config(self):
        import ai_flow.frontend.web_server as web_server
        aiflow_server_config_path = os.path.join(
            os.path.dirname(__file__), "test_aiflow_server.yaml"
        )

        with mock.patch('ai_flow.frontend.web_server.start_web_server') as start_web_server:
            web_server.main(['-c', aiflow_server_config_path])

            config = AIFlowWebServerConfig({'airflow_web_server_uri': 'http://localhost:8080', 'host': '127.0.0.1',
                                            'port': 8000})
            start_web_server.assert_called_once_with('scheduler_class_testing', 'db_uri_testing', config)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import unittest

import sqlalchemy

from ai_flow.store.db.base_model import base
from ai_flow.util import sqlalchemy_db

SQLITE_FILE = 'ai_flow.db'
TEST_URL = 'sqlite:///ai_flow.db'


def create_engine(url):
    return sqlalchemy.create_engine(url)


def get_tables(url):
    return sqlalchemy.inspect(create_engine(url)).get_table_names()


def all_ai_flow_tables_exist(url):
    tables = set(get_tables(url))
    for key in base.metadata.tables.keys():
        if key not in tables:
            return False
    return True


def none_ai_flow_tables_exist(url):
    tables = set(get_tables(url))
    for key in base.metadata.tables.keys():
        if key in tables:
            return False
    return True


class TestSqlalchemyDB(unittest.TestCase):
    def setUp(self) -> None:
        sqlalchemy_db.clear_db(TEST_URL, base.metadata)

    def tearDown(self) -> None:
        if os.path.exists(SQLITE_FILE):
            os.remove(SQLITE_FILE)

    def test_upgrade(self):
        self.assertTrue(none_ai_flow_tables_exist(TEST_URL))
        sqlalchemy_db.upgrade(TEST_URL)
        self.assertTrue(all_ai_flow_tables_exist(TEST_URL))

    def test_upgrade_with_version(self):
        self.assertTrue(none_ai_flow_tables_exist(TEST_URL))
        sqlalchemy_db.upgrade(TEST_URL, 'de1c96ef582a')
        self.assertFalse(all_ai_flow_tables_exist(TEST_URL))
        self.assertTrue(len(get_tables(TEST_URL)) > 0)
        sqlalchemy_db.upgrade(TEST_URL)
        self.assertTrue(all_ai_flow_tables_exist(TEST_URL))

    def test_downgrade(self):
        self.assertTrue(none_ai_flow_tables_exist(TEST_URL))
        sqlalchemy_db.upgrade(TEST_URL)
        self.assertTrue(all_ai_flow_tables_exist(TEST_URL))
        sqlalchemy_db.downgrade(TEST_URL, 'de1c96ef582a')
        self.assertFalse(all_ai_flow_tables_exist(TEST_URL))


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import unittest
import ai_flow.settings

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from ai_flow.cli import cli_parser
from ai_flow.cli.commands import db_command
from ai_flow.settings import get_configuration
from ai_flow.store.db.db_model import SqlProject


class TestCliDb(unittest.TestCase):
    prev_aiflow_home = None

    @classmethod
    def setUpClass(cls):
        cls.parser = cli_parser.get_parser()
        cls.prev_aiflow_home = ai_flow.settings.AIFLOW_HOME
        ai_flow.settings.AIFLOW_HOME = os.path.join(os.path.dirname(__file__), '../')
        cls.config = get_configuration()

    @classmethod
    def tearDownClass(cls) -> None:
        ai_flow.settings.AIFLOW_HOME = cls.prev_aiflow_home

    def _remove_db_file(self):
        if os.path.exists('aiflow.db'):
            os.remove('aiflow.db')

    def setUp(self) -> None:
        self._remove_db_file()

    def tearDown(self) -> None:
        self._remove_db_file()

    def test_cli_db_init(self):
        db_command.init(self.parser.parse_args(['db', 'init']))
        engine = create_engine(self.config.get_db_uri())
        self.assertTrue('project' in engine.table_names())
        self.assertTrue('aiflow_member' in engine.table_names())

    def test_cli_db_reset(self):
        db_command.init(self.parser.parse_args(['db', 'init']))
        engine = create_engine(self.config.get_db_uri())
        Session = scoped_session(
            sessionmaker(autocommit=False,
                         autoflush=False,
                         bind=engine,
                         expire_on_commit=False))
        session = Session()
        project = SqlProject()
        project.name = 'a'
        project.uri = 'a'
        session.add(project)
        session.commit()
        project_result = session.query(SqlProject).all()
        self.assertEqual(1, len(project_result))
        db_command.reset(self.parser.parse_args(['db', 'reset', '-y']))
        project_result = session.query(SqlProject).all()
        self.assertEqual(0, len(project_result))
        session.close()

    def test_cli_db_upgrade(self):
        db_command.upgrade(self.parser.parse_args(['db', 'upgrade', '--version', 'de1c96ef582a']))
        engine = create_engine(self.config.get_db_uri())
        self.assertTrue('project' in engine.table_names())
        self.assertFalse('aiflow_member' in engine.table_names())

    def test_cli_db_downgrade(self):
        db_command.init(self.parser.parse_args(['db', 'init']))
        db_command.downgrade(self.parser.parse_args(['db', 'downgrade', '--version', 'de1c96ef582a']))
        engine = create_engine(self.config.get_db_uri())
        self.assertTrue('project' in engine.table_names())
        self.assertFalse('aiflow_member' in engine.table_names())


if __name__ == '__main__':
    unittest.main()
